/********************************************************************
 Software License Agreement:

 The software supplied herewith by Microchip Technology Incorporated
 (the "Company") for its PIC(R) Microcontroller is intended and
 supplied to you, the Company's customer, for use solely and
 exclusively on Microchip PIC Microcontroller products. The
 software is owned by the Company and/or its supplier, and is
 protected under applicable copyright laws. All rights are reserved.
 Any use in violation of the foregoing restrictions may subject the
 user to criminal sanctions under applicable laws, as well as to
 civil liability for the breach of the terms and conditions of this
 license.

 THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
 WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *******************************************************************/

/** INCLUDES *******************************************************/
#include "usb/usb.h"
#include "usb/usb_device_hid.h"

#include "debug.h"

#include <string.h>
#include <stdio.h>
#include "system.h"
#include "flash.h"
#include "main.h"
#include "patterns.h"


// String for creating debug messages
char debugString[80];

/** VARIABLES ******************************************************/
/* Some processors have a limited range of RAM addresses where the USB module
 * is able to access.  The following section is for those devices.  This section
 * assigns the buffers that need to be used by the USB module into those
 * specific areas.
 */
#if defined(FIXED_ADDRESS_MEMORY)
    #if defined(COMPILER_MPLAB_C18)
        #pragma udata HID_CUSTOM_OUT_DATA_BUFFER = HID_CUSTOM_OUT_DATA_BUFFER_ADDRESS
        unsigned char ReceivedDataBuffer[64];
        #pragma udata HID_CUSTOM_IN_DATA_BUFFER = HID_CUSTOM_IN_DATA_BUFFER_ADDRESS
        unsigned char ToSendDataBuffer[64];
        #pragma udata

    #else defined(__XC8)
        unsigned char ReceivedDataBuffer[64] @ HID_CUSTOM_OUT_DATA_BUFFER_ADDRESS;
        unsigned char ToSendDataBuffer[64] @ HID_CUSTOM_IN_DATA_BUFFER_ADDRESS;
    #endif
#else
    unsigned char ReceivedDataBuffer[64];
    unsigned char ToSendDataBuffer[64];
#endif

volatile USB_HANDLE USBOutHandle;    
volatile USB_HANDLE USBInHandle;

void processUsbCommands();

/** DEFINITIONS ****************************************************/
typedef enum
{
    COMMAND_DEBUG_INFO = 0x10,
    COMMAND_TOGGLE_LED = 0x80,
    COMMAND_GET_BUTTON_STATUS = 0x81,
    COMMAND_GET_LED_STATUS = 0x82,
    COMMAND_SET_LEDS_STATUS = 0x83,
    COMMAND_PATTERN_CHANGE = 0x84,
    COMMAND_STORE_PATTERN = 0x85,
    COMMAND_MULTI_LED_SET = 0x86,
    COMMAND_BLINKIE_INFO = 0x87,
    COMMAND_BLINKIE_SET = 0x88,
    COMMAND_BLINKIE_STATUS = 0x89
} CUSTOM_HID_DEMO_COMMANDS;

/** FUNCTIONS ******************************************************/

/*********************************************************************
* Function: void APP_DeviceCustomHIDInitialize(void);
*
* Overview: Initializes the Custom HID demo code
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceCustomHIDInitialize()
{
    //initialize the variable holding the handle for the last
    // transmission
    USBInHandle = 0;

    //enable the HID endpoint
    USBEnableEndpoint(CUSTOM_DEVICE_HID_EP, USB_IN_ENABLED|USB_OUT_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);

    //Re-arm the OUT endpoint for the next packet
    USBOutHandle = (volatile USB_HANDLE)HIDRxPacket(CUSTOM_DEVICE_HID_EP,(uint8_t*)&ReceivedDataBuffer,64);
}

/*********************************************************************
* Function: void APP_DeviceCustomHIDTasks(void);
*
* Overview: Keeps the Custom HID demo running.
*
* PreCondition: The demo should have been initialized and started via
*   the APP_DeviceCustomHIDInitialize() and APP_DeviceCustomHIDStart() demos
*   respectively.
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceCustomHIDTasks()
{

   // Process USB Commands
   processUsbCommands();

   // Note: Other application specific actions can be placed here

}

extern uint8_t LEDS[8][8];

void erase_flash_mem(uint8_t hi_add, uint8_t low_add)
{
    asm("BCF  INTCON, 7"); // 7 = GIE
    asm("BANKSEL PMCON1");
    PMADRL = low_add;
    PMADRH = hi_add;
    asm("BCF  PMCON1, 6"); // 6=GFGS=0
    asm("BSF  PMCON1, 4"); // 4=FREE=1
    asm("BSF  PMCON1, 2"); // 2=WREN=1
    asm("BSF  INTCON, 7"); // 7=GIE=1

    asm("MOVLW  55h");
    asm("MOVWF  PMCON2");
    asm("MOVLW  0AAh");
    asm("MOVWF  PMCON2");
    asm("BSF    PMCON1, 1"); // 1=WR
    asm("NOP");
    asm("NOP");

    asm("BCF    PMCON1, 2"); // 2=WREN
    asm("BSF    INTCON, 7"); // 7=GIE
}

void write_flash_mem(uint8_t hi_add, uint8_t low_add, char buffer[])
{
    uint8_t i;

//    asm("BCF  INTCON, 7"); // 7 = GIE
    INTCONbits.GIE = 0;
    //    asm("BANKSEL PMCON1");
//    asm("BCF  PMCON1, 6"); // 6=GFGS
//    asm("BCF  PMCON1, 4"); // 4=FREE=0
//    asm("BSF  PMCON1, 2"); // 2=WREN
//    asm("BSF  PMCON1, 5"); // 5=LWLO
    PMCON1bits.CFGS = 0;
    PMCON1bits.FREE = 0;
    PMCON1bits.WREN = 1;
    PMCON1bits.LWLO = 1;

    PMADRL = low_add;
    PMADRH = hi_add;

    // Load the 31 bytes of Write Latches
    for (i=0;i<31;i++) {
        PMDATL = buffer[i];
        PMDATH = 0x34;
        asm("BANKSEL PMCON2");
        asm("MOVLW  55h");
        asm("MOVWF  PMCON2");
        asm("MOVLW  0AAh");
        asm("MOVWF  PMCON2");
        asm("BSF    PMCON1, 1"); // 1=WR
        asm("NOP");
        asm("NOP");
        PMADRL++;
    }
    PMDATL = buffer[31];
    PMDATH = 0x34;
    // Flash the 32 bytes of data

    PMCON1bits.LWLO = 0;
    asm("BANKSEL PMCON2");
    // asm("BCF  PMCON1, 5"); //5=LWLO

    asm("MOVLW  55h");
    asm("MOVWF  PMCON2");
    asm("MOVLW  0AAh");
    asm("MOVWF  PMCON2");
    asm("BSF    PMCON1, 1"); // 1=WR
    asm("NOP");
    asm("NOP");

    //
//    asm("BCF    PMCON1, 2"); //2=WREN
    PMCON1bits.WREN = 0;
//    asm("BSF    INTCON, 7"); // 7=GIE
    INTCONbits.GIE = 1;
}

void store_settings(uint8_t demo_mode, uint8_t user_msg_size, uint8_t user_id, uint8_t plockout[4]) {
    char buffer[32];
    
    buffer[0] = demo_mode;
    buffer[1] = user_msg_size;
    buffer[2] = user_id;
    
    erase_flash_mem(0x1f,0x00);
    write_flash_mem(0x1f,0x00,buffer);
    
}

void processUsbCommands()
{
    uint8_t x,y,i;
    uint8_t count;
    //Check if we have received an OUT data packet from the host
    if(HIDRxHandleBusy(USBOutHandle) == false)
    {   
        //We just received a packet of data from the USB host.
        //Check the first uint8_t of the packet to see what command the host
        //application software wants us to fulfill.
        switch(ReceivedDataBuffer[0])				//Look at the data the host sent, to see what kind of application specific command it sent.
        {
            case COMMAND_DEBUG_INFO:  //Debug information request from host
                // Copy any waiting debug text to the send data buffer
		copyDebugToSendBuffer((uint8_t*)&ToSendDataBuffer[0]);

		// Transmit the response to the host
                if(!HIDTxHandleBusy(USBInHandle))
		{
		   USBInHandle = HIDTxPacket(CUSTOM_DEVICE_HID_EP,(uint8_t*)&ToSendDataBuffer[0],64);
		}
                break;
            case COMMAND_TOGGLE_LED:  //Toggle LEDs command
//                LED_Toggle(LED_USB_DEVICE_HID_CUSTOM);
                run = !run;
                break;
            case COMMAND_GET_BUTTON_STATUS:  //Get push button state
                //Check to make sure the endpoint/buffer is free before we modify the contents
                if(!HIDTxHandleBusy(USBInHandle))
                {
                    ToSendDataBuffer[0] = 0x81;				//Echo back to the host PC the command we are fulfilling in the first uint8_t.  In this case, the Get Pushbutton State command.
                    if(sw0 == false)	//pushbutton not pressed, pull up resistor on circuit board is pulling the PORT pin high
                    {
                            ToSendDataBuffer[1] = 0x01;
                    }
                    else									//sw3 must be == 0, pushbutton is pressed and overpowering the pull up resistor
                    {
                            ToSendDataBuffer[1] = 0x00;
                    }
                    //Prepare the USB module to send the data packet to the host
                    USBInHandle = HIDTxPacket(CUSTOM_DEVICE_HID_EP, (uint8_t*)&ToSendDataBuffer[0],64);
                }
                break;

            case COMMAND_GET_LED_STATUS:  //Get push button state
                //Check to make sure the endpoint/buffer is free before we modify the contents
                if(!HIDTxHandleBusy(USBInHandle))
                {
                    ToSendDataBuffer[0] = 0x82;				//Echo back to the host PC the command we are fulfilling in the first uint8_t.  In this case, the Get Pushbutton State command.
                    if(sw0 == false)	//pushbutton not pressed, pull up resistor on circuit board is pulling the PORT pin high
                    {
                            ToSendDataBuffer[1] = 0x01;
                    }
                    else									//sw3 must be == 0, pushbutton is pressed and overpowering the pull up resistor
                    {
                            ToSendDataBuffer[1] = 0x00;
                    }
                    //Prepare the USB module to send the data packet to the host
                    USBInHandle = HIDTxPacket(CUSTOM_DEVICE_HID_EP, (uint8_t*)&ToSendDataBuffer[0],64);
                }
                break;

            case COMMAND_SET_LEDS_STATUS:
                x = ReceivedDataBuffer[4]>>3; // 7-26-2014 DDF not working?
                y = ReceivedDataBuffer[4]%8;
//                sprintf(debugString, "%d %2X", y, patterns[11][y]);
                sprintf(debugString, "LEDS[%d][%d] = %d -> %d",x,y,LEDS[x][y],ReceivedDataBuffer[1]);
	            debugOut(debugString);
                LEDS[x][y] = ReceivedDataBuffer[1];
                break;
            case COMMAND_PATTERN_CHANGE:  // pattern change command
				// process an RBG matrix command
		        if (ReceivedDataBuffer[1] == 0) {
                   next_pattern();
                } else if (ReceivedDataBuffer[1] == 1) {
                   back_pattern();
                }
            	break;
            case COMMAND_STORE_PATTERN:
                  for (i=1;ReceivedDataBuffer[i]!=0x00;i++) {}
                  user_msg_size = i;
                  erase_flash_mem(0x1f,0x20);
                  write_flash_mem(0x1f,0x20,&ReceivedDataBuffer[1]);
                  if (user_msg_size > 31) {
                     erase_flash_mem(0x1f,0x40);
                     write_flash_mem(0x1f,0x40,&ReceivedDataBuffer[33]); //got 77 with 32
                  }
                  sprintf(debugString,"%d %s",user_msg_size, &ReceivedDataBuffer[4]);
                  debugOut(debugString);
                  store_settings(demo_mode, user_msg_size, user_id, plockout);
                  sprintf(debugString,"%d %d %d",user_msg_size, demo_mode, user_id);
	              debugOut(debugString);
                  if (p_table == 0) p_count = 0;  // If we are playing pattern 0, restart it.
                break;
            case COMMAND_MULTI_LED_SET:  // multi RGB LED set command

                for  (count=0; count < ReceivedDataBuffer[1];count++) {
                   x = ReceivedDataBuffer[3+count*3]>>3; // 7-26-2014 DDF not working?
                   y = ReceivedDataBuffer[3+count*3]%8;
                   LEDS[x][y] = ReceivedDataBuffer[2+count*3] & 0x0f;
                }
                break;
            case COMMAND_BLINKIE_INFO:  // Blinkie Info

                ToSendDataBuffer[0] = COMMAND_BLINKIE_INFO;
                ToSendDataBuffer[1] = 0x06;   // 8x7 mono
                ToSendDataBuffer[2] = 0x01;  // Version 1
                ToSendDataBuffer[3] = 0x00;  // LEDs installed
                ToSendDataBuffer[4] = user_id; 
                ToSendDataBuffer[5] = 0x00;  // Pattern on SD 
                ToSendDataBuffer[6] = plockout[0];
                ToSendDataBuffer[7] = plockout[1];
                ToSendDataBuffer[8] = plockout[2];
                ToSendDataBuffer[9] = plockout[3];
                
				// Transmit the response to the host
                if(!HIDTxHandleBusy(USBInHandle))
				{
                    USBInHandle = HIDTxPacket(CUSTOM_DEVICE_HID_EP, (uint8_t*)&ToSendDataBuffer[0],64);
				}
            	break;
            case COMMAND_BLINKIE_SET:
                // LED_type = ReceivedDataBuffer[1];
                user_id = ReceivedDataBuffer[2];
                demo_mode = ReceivedDataBuffer[3];
                plockout[0] = ReceivedDataBuffer[4];
                plockout[1] = ReceivedDataBuffer[5];
                plockout[2] = ReceivedDataBuffer[6];
                plockout[3] = ReceivedDataBuffer[7];
                store_settings(demo_mode, user_msg_size, user_id, plockout);
                break;
            case COMMAND_BLINKIE_STATUS:
                ToSendDataBuffer[0] = COMMAND_BLINKIE_STATUS;
                ToSendDataBuffer[1] = (sw0)? 0x00 : 0x01; //Push Button status
                ToSendDataBuffer[2] = demo_mode | run;
                ToSendDataBuffer[3] = p_table;  // pattern #
                ToSendDataBuffer[4] = 'u';  // pattern name 0
                ToSendDataBuffer[5] = 'n';  // pattern name 1
                ToSendDataBuffer[6] = 'd';  // pattern name 2
                ToSendDataBuffer[7] = 'e';  // pattern name 3
                ToSendDataBuffer[8] = 'f';  // pattern name 4
                ToSendDataBuffer[9] = 'i';  // pattern name 5
                ToSendDataBuffer[10] = 'n';  // pattern name 6
                ToSendDataBuffer[11] = 'e';  // pattern name 7

                // Transmit the response to the host
                if(!HIDTxHandleBusy(USBInHandle))
				{
					USBInHandle = HIDTxPacket(CUSTOM_DEVICE_HID_EP, (uint8_t*)&ToSendDataBuffer[0],64);
				}
            	break;

            default:   // unknown command received
                break;
        }
        //Re-arm the OUT endpoint, so we can receive the next OUT data packet 
        //that the host may try to send us.
        USBOutHandle = HIDRxPacket(CUSTOM_DEVICE_HID_EP, (uint8_t*)&ReceivedDataBuffer, 64);
    }
}